package com.devbeh.managers;


import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.time.Duration;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicLong;

import com.devbeh.ActionClass;
import com.devbeh.Actor;
import com.devbeh.DetectionClass;
import com.devbeh.MetricService;
import com.devbeh.Utils;
import com.devbeh.commondata.DevBehDetections;
import com.devbeh.commondata.Empty;
import com.devbeh.commondata.KBEntry;
import com.devbeh.commondata.KBObject;
import com.devbeh.commondata.KBPoint2d;
import com.devbeh.commondata.KBPoint3d;
import com.devbeh.commondata.KBServiceCommand;
import com.devbeh.commondata.KBServiceResponse;
import com.devbeh.data.Action;
import com.devbeh.data.DetectionEntryOrBuilder;
import com.devbeh.data.DistanceEstimationEntry;
import com.devbeh.data.PerceptionResponse;
import com.devbeh.data.PerceptionResponseOrBuilder;
import com.devbeh.data.Point3d;
import com.devbeh.data.Point3dOrBuilder;
import com.devbeh.data.TrackingEntryOrBuilder;
import com.devbeh.kafka.DevBehDetectionsSerializer;
import com.devbeh.kafka.KBEntryDeserializer;
import com.devbeh.kafka.KBEntrySerializer;
import com.devbeh.kafka.KafkaUtils;
import org.apache.commons.io.IOUtils;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.OffsetAndTimestamp;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class KBManager extends Actor implements TrackingConsumer, DistanceDataConsumer, ActionsConsumer
{
    private static final Logger LOGGER = LoggerFactory.getLogger(KBManager.class);

    /**
     * Счетчик ошибок.
     */
    private final AtomicLong errorCounter = new AtomicLong();

    /**
     * Отображает действие на идентификатор кадра начала действия.
     */
    private final Map<Action, Long> activeActions = new HashMap<>();

    /**
     * Имя данного менеджера.
     */
    private final String managerName;

    /**
     * Адрес сервера Kafka.
     */
    private final String bootstrapServer;

    /**
     * Имя пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String username;

    /**
     * Пароль пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String password;

    /**
     * Идентификатор клиента (и группы) Kafka, используемый для работы с топиками состояния и результатов
     * детектирования.
     */
    private final String clientId;

    /**
     * Топик для хранения атомарных знаний, детектированных по кадкам видеозаписи.
     */
    private final TopicPartition kbStateTopicPartition;

    /**
     * Топик для хранения результатов детектирования девиантного поведения людей.
     */
    private final TopicPartition devBehDetectionsTopicPartition;

    /**
     * Размер интервала времени, который охватывает база знаний, в секундах.
     */
    private final int kbHistorySec;

    /**
     * Период, с которым осуществляется логический вывод по базе знаний для детектирования девиантного поведения,
     * в миллисекундах.
     */
    private final int inferencePeriodSec;

    /**
     * Период, в течение которого после детектирования определенного инстанса девиантного поведения, аналогичные
     * инстансы будут проигнорированы (инстансы одного типа девиантного поведения с одними действующими лицами).
     */
    private final int devBehIgnorePeriodSec;

    /**
     * Предоставляет fps обрабатываемого видеопотока. Используется для создания результирующего видео.
     */
    private final Future<Double> fpsFuture;

    /**
     * Путь к jar компонента-обертки над базой знаний.
     */
    private final String kbWorkerJarPath;

    /**
     * Значение параметров Xms Xmx при запуске компонента-обертки над базой знаний.
     */
    private final String kbWorkerMemory;

    /**
     * Путь к директории, хранящей библиотечный файл libjpl.so.
     */
    private final String libjplJavaLibraryPath;

    /**
     * Путь к файлу, содержащему общие статические правила, описывающие девиантное поведение.
     */
    private final String generalKbRulesFilePath;

    /**
     * Путь к файлу, содержащему специфичные для данного видеопотока правила, или пустая строка.
     */
    private final String cameraKbRulesFilePath;

    /**
     * Максимальная длительность действия в секундах, по достижению которой будет отправлено сообщение о завершении
     * действия.
     */
    private final int maxActionDurationSec;

    /**
     * Используется для классификации ребенка по размеру скелета.
     */
    private final int maxChildSkeletonLengthMm;

    /**
     * Используется для остановки конвейера обработки видео в случае критичной ошибки.
     */
    private final Runnable pipelineStopper;

    /**
     * Используется для публикации метрик.
     */
    private final MetricService metricService;

    /**
     * Имя метрики последнего обработанного идентификатора кадра.
     */
    private final String lastFrameIdMetricName;

    /**
     * Имя метрики количества ошибок.
     */
    private final String totalErrorsMetricName;

    /**
     * После инициализации содержит идентификатор первого кадра, который требуется данному компоненту для обработки.
     */
    private final CompletableFuture<Long> initialLastConsumedFrameIdFuture;

    /**
     * Очередь результатов оценки отслеживания объектов.
     */
    private final BlockingQueue<PerceptionResponseOrBuilder> trackingQueue;

    /**
     * Очередь результатов оценки положения объектов в координатах камеры.
     */
    private final BlockingQueue<PerceptionResponse> distanceQueue;

    /**
     * Очередь результатов детектирования действий людей.
     */
    private final BlockingQueue<PerceptionResponseOrBuilder> actionsQueue;

    /**
     * Используется для записи в Kafka детектированных атомарных знаний.
     */
    private KafkaProducer<?, KBEntry> kbStateProducer;

    /**
     * Используется для записи в Kafka результатов детектирования девиантного поведения.
     */
    private KafkaProducer<?, DevBehDetections> devBehDetectionProducer;

    /**
     * Размер интервала времени, который охватывает база знаний, в кадрах.
     */
    private int kbHistoryFrames;

    /**
     * Период, с которым осуществляется логический вывод по базе знаний, в кадрах.
     */
    private int inferencePeriodFrames;

    /**
     * Период, в течение которого будут игнорироваться аналогичные детектируемые инстансы девиантного поведения.
     */
    private int devBehIgnorePeriodFrames;

    /**
     * Максимальная длительность действия в кадрах.
     */
    private int maxActionDurationFrames;

    /**
     * Процесс-обертка над базой знаний или null.
     */
    private Process kbWorkerProcess;

    /**
     * Входной стрим процесса-обертки над базой знаний для отправки команд или null.
     */
    private InputStream kbWorkerIn;

    /**
     * Выходной стрим процесса-обертки над базой знаний для получения результатов выполнения команд или null.
     */
    private OutputStream kbWorkerOut;

    /**
     * Выходной стрим ошибок процесса-обертки над базой знаний или null.
     */
    private Reader kbWorkerErr;

    /**
     * Результат отслеживания для текущего обрабатываемого кадра или null.
     */
    private PerceptionResponseOrBuilder currentTrackingResult;

    /**
     * Результат оценки положения объектов в системе координат камеры для текущего обрабатываемого кадра или null.
     */
    private PerceptionResponse currentDistanceResult;

    /**
     * Результат детектирования действий людей для текущего обрабатываемого кадра.
     */
    private PerceptionResponseOrBuilder currentActionsResult;

    /**
     * Идентификатор последнего отправленного на обработку кадра. Используется для подбора идентификатора
     * завершающего кадра.
     */
    private long lastProcessedFrameId;

    /**
     * Идентификатор последнего кадра, на котором был осуществлен логический вывод по базе знаний.
     */
    private long lastInferenceFrameId = -1;


    public KBManager(String videoName, int videoId, String bootstrapServer, String username, String password,
            String kbStateTopicPrefix, String devBehDetectionsTopicPrefix, int kbHistorySec, int inferencePeriodSec,
            int devBehIgnorePeriodSec, Future<Double> fpsFuture, int maxQueuedRequests, String kbWorkerJarPath,
            String kbWorkerMemory, String libjplJavaLibraryPath, String generalKbRulesFilePath,
            String cameraKbRulesFilePath, int maxActionDurationSec, int maxChildSkeletonLengthMm,
            Runnable pipelineStopper, MetricService metricService)
    {
        this.managerName = "KBManager-" + videoName;
        this.bootstrapServer = bootstrapServer;
        this.username = username;
        this.password = password;
        this.clientId = "kb-manager-" + videoId;
        this.kbStateTopicPartition = new TopicPartition(kbStateTopicPrefix + videoId, 0);
        this.devBehDetectionsTopicPartition = new TopicPartition(devBehDetectionsTopicPrefix + videoId, 0);
        this.kbHistorySec = kbHistorySec;
        this.inferencePeriodSec = inferencePeriodSec;
        this.devBehIgnorePeriodSec = devBehIgnorePeriodSec;
        this.fpsFuture = fpsFuture;
        this.kbWorkerJarPath = kbWorkerJarPath;
        this.kbWorkerMemory = kbWorkerMemory;
        this.libjplJavaLibraryPath = libjplJavaLibraryPath;
        this.generalKbRulesFilePath = generalKbRulesFilePath;
        this.cameraKbRulesFilePath = Utils.isEmpty(cameraKbRulesFilePath) ? "" : cameraKbRulesFilePath;
        this.maxActionDurationSec = maxActionDurationSec;
        this.maxChildSkeletonLengthMm = maxChildSkeletonLengthMm;
        this.pipelineStopper = pipelineStopper;
        this.metricService = metricService;
        String metricsPrefix = videoName + ".kb.";
        this.lastFrameIdMetricName = metricsPrefix + "lastFrameId";
        this.totalErrorsMetricName = metricsPrefix + "totalErrors";
        this.initialLastConsumedFrameIdFuture = new CompletableFuture<>();
        this.trackingQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.distanceQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.actionsQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        metricService.submit(totalErrorsMetricName, 0);
    }


    @Override
    public long getInitialLastConsumedFrameId() throws InterruptedException
    {
        try
        {
            return initialLastConsumedFrameIdFuture.get();
        }
        catch (ExecutionException e)
        {
            // Shouldn't happen!
            throw new RuntimeException(e.getCause());
        }
    }


    @Override
    public void putTracking(PerceptionResponseOrBuilder tracking) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            trackingQueue.put(tracking);
        }
    }


    @Override
    public void putDistances(PerceptionResponse distances) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            distanceQueue.put(distances);
        }
    }


    @Override
    public void putActions(PerceptionResponseOrBuilder actions) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            actionsQueue.put(actions);
        }
    }


    @Override
    protected void doFirst()
    {
        double fps;
        try
        {
            fps = fpsFuture.get();
        }
        catch (Exception e)
        {
            if (!getDoStop())
            {
                // Shouldn't happen!
                handleCriticalError("Exception occurred during fps initialization", e);
            }
            return;
        }

        kbHistoryFrames = (int) (fps * kbHistorySec);
        inferencePeriodFrames = (int) (fps * inferencePeriodSec);
        devBehIgnorePeriodFrames = (int) (fps * devBehIgnorePeriodSec);
        maxActionDurationFrames = (int) (fps * maxActionDurationSec);

        initKb(fps);
        loadKbHistory();

        if (getDoStop())
        {
            return;
        }

        kbStateProducer = KafkaUtils.buildKafkaProducer(bootstrapServer, username, password, clientId + "-state",
                new KBEntrySerializer());
        devBehDetectionProducer = KafkaUtils.buildKafkaProducer(bootstrapServer, username, password, clientId,
                new DevBehDetectionsSerializer());
    }


    /**
     * Загружает историю в базу знаний.
     */
    private void loadKbHistory()
    {
        KafkaConsumer<?, KBEntry> kbStateConsumer = KafkaUtils.buildKafkaConsumer(bootstrapServer, username,
                password, clientId + "-state", clientId + "-state", kbStateTopicPartition,
                new KBEntryDeserializer());

        Collection<TopicPartition> kbStateTPCollection = Collections.singleton(kbStateTopicPartition);
        long beginningOffset = kbStateConsumer.beginningOffsets(kbStateTPCollection).get(kbStateTopicPartition);
        long endOffset = kbStateConsumer.endOffsets(kbStateTPCollection).get(kbStateTopicPartition);

        if (beginningOffset == endOffset)
        {
            LOGGER.info("State history is empty (topic is empty) [{}]. Starting from the beginning", managerName);
            initialLastConsumedFrameIdFuture.complete(-1L);
            lastProcessedFrameId = -1;
            kbStateConsumer.close();
            return;
        }

        // Определяем время последней записи в топике

        kbStateConsumer.seek(kbStateTopicPartition, endOffset-1);
        long lastTimestamp;
        while (true)
        {
            if (getDoStop())
            {
                kbStateConsumer.close();
                return;
            }

            ConsumerRecords<?, KBEntry> records = kbStateConsumer.poll(Duration.ofMillis(100));
            if (!records.isEmpty())
            {
                ConsumerRecord<?, KBEntry> r = records.iterator().next();
                lastTimestamp = r.timestamp();
                break;
            }
        }

        // Загружаем историю

        long initialKbStateFrameId = Math.max(0, lastTimestamp - kbHistoryFrames);
        OffsetAndTimestamp initialKbStateOffsetAndFrameId = kbStateConsumer.offsetsForTimes(
                Map.of(kbStateTopicPartition, initialKbStateFrameId)).get(kbStateTopicPartition);
        if (initialKbStateOffsetAndFrameId == null)
        {
            // Shouldn't happen!
            handleCriticalError("Unexpectedly the last kb state record is missing. Manager: " + managerName,
                    null);
            kbStateConsumer.close();
            return;
        }

        kbStateConsumer.seek(kbStateTopicPartition, initialKbStateOffsetAndFrameId.offset());
        long currentFrameId = initialKbStateOffsetAndFrameId.timestamp() - 1;

        loadKbHistoryCycle:
        while (!getDoStop())
        {
            for (ConsumerRecord<?, KBEntry> r : kbStateConsumer.poll(Duration.ofMillis(100)))
            {
                if (currentFrameId < r.timestamp())
                {
                    if (r.value().getFrameId() == -1)
                    {
                        currentFrameId = Long.MAX_VALUE;
                        break loadKbHistoryCycle;
                    }
                    else
                    {
                        currentFrameId = r.timestamp();
                        addFrameKnowledge(r.value());
                    }
                }

                if (r.offset() == endOffset-1)
                {
                    break loadKbHistoryCycle;
                }
            }
        }
        kbStateConsumer.close();

        if (getDoStop())
        {
            return;
        }

        initialLastConsumedFrameIdFuture.complete(currentFrameId);
        lastProcessedFrameId = currentFrameId;
        if (currentFrameId == Long.MAX_VALUE)
        {
            stop();
            submitLastFrameId(-1);
        }
    }


    /**
     * Обрабатывает критическую ошибку: логирует, увеличивает счетчик ошибок и останавливает pipeline обработки видео.
     *
     * @param errorMsg описание ошибки
     * @param e возникшее исключение или null
     */
    protected void handleCriticalError(String errorMsg, Exception e)
    {
        if (e == null)
        {
            LOGGER.error(errorMsg);
        }
        else
        {
            LOGGER.error(errorMsg, e);
        }
        incrementErrorCounter();
        pipelineStopper.run();
    }


    /**
     * Увеличивает счетчик ошибок и обновляет соответствущую метрику.
     */
    private void incrementErrorCounter()
    {
        metricService.submit(totalErrorsMetricName, errorCounter.incrementAndGet());
    }


    /**
     * Отображает в метриках последний обработанный кадр видеозаписи.
     */
    private void submitLastFrameId(long frameId)
    {
        metricService.submit(lastFrameIdMetricName, frameId);
    }


    @Override
    protected void process() throws Exception
    {
        if (!isAllFrameDataAvailable())
        {
            ensureNextFrameData();
        }

        KBEntry.Builder kbEntryBuilder = KBEntry.newBuilder()
                .setFrameId(currentTrackingResult.getFrameId());
        long frameId;
        if (currentTrackingResult.getFrameId() == -1)
        {
            frameId = lastProcessedFrameId + 1;
        }
        else
        {
            frameId = currentTrackingResult.getFrameId();
            lastProcessedFrameId = frameId;
            buildKbEntry(kbEntryBuilder);
        }
        KBEntry kbEntry = kbEntryBuilder.build();

        // Цикл вызова process может застрять на переотправке данной записи, если ее размер превышает максимум.
        // Считаем, что все отправляемые сообщения будут укладываться в ограничение.
        kbStateProducer.send(new ProducerRecord<>(kbStateTopicPartition.topic(), kbStateTopicPartition.partition(),
                frameId, null, kbEntry)).get();

        if (currentTrackingResult.getFrameId() == -1)
        {
            // Выполняем логический вывод в последний раз (с предыдущим кадром)
            runInference();
            // Отправляем завершающее сообщение в топик детектированных экземпляров девиантного поведения
            sendDevBehDetectionsToKafka(DevBehDetections.newBuilder().setFrameId(-1).build(), frameId);
            // Сообщаем о завершении обработки видео
            submitLastFrameId(-1);
            // Завершаем работу менеджера
            stop();
            return;
        }

        addFrameKnowledge(kbEntry);

        if (lastInferenceFrameId + inferencePeriodFrames < currentTrackingResult.getFrameId())
        {
            lastInferenceFrameId = currentTrackingResult.getFrameId();
            runInference();
        }

        submitLastFrameId(currentTrackingResult.getFrameId());

        clearFrameData();
    }


    @SuppressWarnings("BooleanMethodIsAlwaysInverted")
    private boolean isAllFrameDataAvailable()
    {
        return currentTrackingResult != null && currentDistanceResult != null && currentActionsResult != null
                && currentTrackingResult.getFrameId() == currentDistanceResult.getFrameId()
                && currentTrackingResult.getFrameId() == currentActionsResult.getFrameId();
    }


    private void ensureNextFrameData() throws InterruptedException
    {
        while (!isAllFrameDataAvailable())
        {
            if (currentTrackingResult == null)
            {
                currentTrackingResult = trackingQueue.take();
            }

            if (currentDistanceResult == null)
            {
                currentDistanceResult = distanceQueue.take();
            }

            if (currentActionsResult == null)
            {
                currentActionsResult = actionsQueue.take();
            }

            if (currentTrackingResult.getFrameId() != currentDistanceResult.getFrameId())
            {
                if (currentTrackingResult.getFrameId() == -1 || currentDistanceResult.getFrameId() != -1
                        && currentDistanceResult.getFrameId() < currentTrackingResult.getFrameId())
                {
                    LOGGER.warn("Missing a tracking result with id {} [{}]", currentDistanceResult.getFrameId(),
                            managerName);
                    currentDistanceResult = distanceQueue.take();
                }
                else
                {
                    LOGGER.warn("Missing a distance result with frame id {} [{}]", currentTrackingResult.getFrameId(),
                            managerName);
                    currentTrackingResult = trackingQueue.take();
                }
            }

            if (currentTrackingResult.getFrameId() != currentActionsResult.getFrameId())
            {
                if (currentTrackingResult.getFrameId() == -1 || currentActionsResult.getFrameId() != -1
                        && currentActionsResult.getFrameId() < currentTrackingResult.getFrameId())
                {
                    LOGGER.warn("Missing a tracking with id {} [{}]", currentDistanceResult.getFrameId(), managerName);
                    currentActionsResult = actionsQueue.take();
                }
                else
                {
                    LOGGER.warn("Missing an actions result with frame id {} [{}]", currentTrackingResult.getFrameId(),
                            managerName);
                    currentTrackingResult = trackingQueue.take();
                }
            }
        }
    }


    private void clearFrameData()
    {
        currentTrackingResult = null;
        currentDistanceResult = null;
        currentActionsResult = null;
    }


    /**
     * Заполняет объект знаниями, полученными из результатов анализа кадра.
     *
     * @param kbEntryBuilder объект для записи знаний, относящихся к данному кадру
     */
    private void buildKbEntry(KBEntry.Builder kbEntryBuilder)
    {
        for (int i = 0; i < currentTrackingResult.getTrackingOrBuilder().getEntriesCount(); i++)
        {
            TrackingEntryOrBuilder te = currentTrackingResult.getTrackingOrBuilder().getEntries(i);
            DetectionEntryOrBuilder de = te.getDetection();
            DistanceEstimationEntry dist = currentDistanceResult.getDistance().getEntries(i);

            KBObject.Builder kbObjectBuilder = kbEntryBuilder.addKnowledgeBuilder().getKbObjectBuilder();
            kbObjectBuilder.setClassId(de.getClassId());
            kbObjectBuilder.setGlobalId(te.getId());
            kbObjectBuilder.getBoundingBoxBuilder()
                    .setTopLeft(KBPoint2d.newBuilder().setX(de.getBoxTopLeftX()).setY(de.getBoxTopLeftY()))
                    .setBottomRight(KBPoint2d.newBuilder().setX(de.getBoxBottomRightX()).setY(de.getBoxBottomRightY()));

            Point3dOrBuilder position;
            DetectionClass cl = DetectionClass.findById(de.getClassId());
            if (cl == DetectionClass.PERSON)
            {
                position = dist.getPoints(0);
                for (Point3d p : dist.getPointsList())
                {
                    // На данный момент для упрощения предполагаем, что точка положения объекта это нижняя точка
                    // в кадре
                    if (position.getY() < p.getY())
                    {
                        position = p;
                    }
                }

                // Добавляем вектор направления головы

                Point3d thorax = dist.getPoints(17);
                Point3d nose = dist.getPoints(9);

                // Вектор от грудной клетки до носа человека
                float thoraxNoseX = nose.getX() - thorax.getX();
                float thoraxNoseY = nose.getY() - thorax.getY();
                float thoraxNoseZ = nose.getZ() - thorax.getZ();

                float planeX = currentDistanceResult.getDistance().getNormalVector().getX();
                float planeY = currentDistanceResult.getDistance().getNormalVector().getY();
                float planeZ = currentDistanceResult.getDistance().getNormalVector().getZ();

                float projCoef = (thoraxNoseX * planeX + thoraxNoseY * planeY + thoraxNoseZ * planeZ) /
                        (planeX * planeX + planeY * planeY + planeZ * planeZ);

                float lookX = thoraxNoseX - projCoef * planeX;
                float lookY = thoraxNoseY - projCoef * planeY;
                float lookZ = thoraxNoseZ - projCoef * planeZ;

                kbObjectBuilder.setLookDirection(KBPoint3d.newBuilder().setX(lookX).setY(lookY).setZ(lookZ));

                // Добавляем признак, является ли человек ребенком
                Optional<Integer> skeletonLength = calculateSkeletonLength(dist.getPointsList());
                skeletonLength.ifPresent(len -> kbObjectBuilder.setChild(len <= maxChildSkeletonLengthMm));
            }
            else
            {
                float x = 0, y = 0, z = 0;
                for (Point3d p : dist.getPointsList())
                {
                    x += p.getX();
                    y += p.getY();
                    z += p.getZ();
                }
                x = x / dist.getPointsCount();
                y = y / dist.getPointsCount();
                z = z / dist.getPointsCount();
                position = Point3d.newBuilder().setX(x).setY(y).setZ(z);
            }

            // Добавляем положение в пространстве
            kbObjectBuilder.getPositionBuilder().setX(position.getX()).setY(position.getY()).setZ(position.getZ());

            // Добавляем объемное положение для машин
            if (cl == DetectionClass.CAR || cl == DetectionClass.BICYCLE || cl == DetectionClass.MOTORCYCLE
                    || cl == DetectionClass.BUS || cl == DetectionClass.TRUCK)
            {
                for (Point3d p : dist.getPointsList())
                {
                    kbObjectBuilder.addVolumetricPositionBuilder().setX(p.getX()).setY(p.getY()).setZ(p.getZ());
                }
            }
        }

        // Записываем знания о действиях

        int latencyFrames = currentActionsResult.getActions().getLatencyFrames();
        Set<Action> actions = new HashSet<>();
        for (Action action : currentActionsResult.getActions().getEntriesList())
        {
            ActionClass ac = ActionClass.findById(action.getAction());
            if (ac == null)
            {
                continue;
            }

            // На данный момент некоторые классы детектируемых действий приравниваются к одному типу
            switch (ac)
            {
                case GIVING_OBJECT:
                case EXCHANGE_THINGS:
                    action = buildAction(ActionClass.EXCHANGE_THINGS, action.getFirstPersonId(),
                            action.getSecondPersonId());
                    break;
                case TOUCH_POCKET:
                    action = buildAction(ActionClass.TOUCH_POCKET, action.getFirstPersonId(),
                            action.getSecondPersonId());
                    break;
                case SIT_DOWN:
                case SQUAT_DOWN:
                    action = buildAction(ActionClass.SQUAT_DOWN, action.getFirstPersonId(), 0);
                    break;
                case FALLING_DOWN:
                    action = buildAction(ActionClass.FALLING_DOWN, action.getFirstPersonId(), 0);
                    break;
                case THROW:
                    action = buildAction(ActionClass.THROW, action.getFirstPersonId(), 0);
                    break;
                case PUNCH_SLAP:
                case HIT_WITH_OBJECT:
                case ARM_SWIGS:
                    action = buildAction(ActionClass.ARM_SWIGS, action.getFirstPersonId(), 0);
                    break;
                case SIDE_KICK:
                case BUT_KICKS:
                case KICKING:
                case KICKING_SOMETHING:
                    action = buildAction(ActionClass.KICKING_SOMETHING, action.getFirstPersonId(), 0);
                    break;
                default:
                    action = null;
            }

            if (action == null)
            {
                continue;
            }

            actions.add(action);
            Long actionStartFrameId = activeActions.get(action);
            if (actionStartFrameId == null)
            {
                addAction(action, latencyFrames, true, kbEntryBuilder);
                activeActions.put(action, currentTrackingResult.getFrameId());
            }
            else
            {
                if (currentTrackingResult.getFrameId() - actionStartFrameId > maxActionDurationFrames)
                {
                    // Закрываем устаревшее действие. Оно будет переоткрыто на следующем кадре, если действие продолжится
                    addAction(action, latencyFrames, false, kbEntryBuilder);
                    activeActions.remove(action);
                }
            }
        }

        // Завершаем действия, которые не присутствуют в данном кадре
        Set<Action> finishedActions = new HashSet<>(activeActions.keySet());
        finishedActions.removeAll(actions);
        finishedActions.forEach(a -> addAction(a, latencyFrames, false, kbEntryBuilder));

        activeActions.keySet().retainAll(actions);
    }


    /**
     * Вычисляет длину скелета человека.
     *
     * @param skeleton ключевые точки тела человека
     * @return длина скелета человека, если переданное количество точек равно 18
     */
    private static Optional<Integer> calculateSkeletonLength(List<Point3d> skeleton)
    {
        if (skeleton.size() != 18)
        {
            return Optional.empty();
        }

        // Скелет:
        // 0 <- 1 <- 2 <- 3
        // 0 <- 4 <- 5 <- 6
        // 0 <- 17
        // 17 <- 11 <- 12 <- 13
        // 17 <- 14 <- 15 <- 16
        double skeletonLength = calculateDistance(skeleton.get(0), skeleton.get(1))
                + calculateDistance(skeleton.get(1), skeleton.get(2))
                + calculateDistance(skeleton.get(2), skeleton.get(3))
                + calculateDistance(skeleton.get(0), skeleton.get(4))
                + calculateDistance(skeleton.get(4), skeleton.get(5))
                + calculateDistance(skeleton.get(5), skeleton.get(6))
                + calculateDistance(skeleton.get(0), skeleton.get(17))
                + calculateDistance(skeleton.get(17), skeleton.get(11))
                + calculateDistance(skeleton.get(11), skeleton.get(12))
                + calculateDistance(skeleton.get(12), skeleton.get(13))
                + calculateDistance(skeleton.get(17), skeleton.get(14))
                + calculateDistance(skeleton.get(14), skeleton.get(15))
                + calculateDistance(skeleton.get(15), skeleton.get(16));

        return Optional.of((int) skeletonLength);
    }


    /**
     * Вычисляет расстояние между двумя точками.
     */
    private static double calculateDistance(Point3d p1, Point3d p2)
    {
        double x = (p1.getX() - p2.getX());
        double y = (p1.getY() - p2.getY());
        double z = (p1.getZ() - p2.getZ());
        return Math.sqrt(x*x + y*y + z*z);
    }


    private static Action buildAction(ActionClass actionClass, long firstPerson, long secondPerson)
    {
        Action.Builder ab = Action.newBuilder().setAction(actionClass.getId()).setFirstPersonId(firstPerson);
        if (secondPerson > 0)
        {
            ab.setSecondPersonId(secondPerson);
        }
        return ab.build();
    }


    private static void addAction(Action action, int latencyFrames, boolean started, KBEntry.Builder kbEntryBuilder)
    {
        kbEntryBuilder.addKnowledgeBuilder().getKbActionBuilder()
                .setAction(action.getAction())
                .setFirstPersonId(action.getFirstPersonId())
                .setSecondPersonId(action.getSecondPersonId())
                .setLatencyFrames(latencyFrames)
                .setStarted(started);
    }


    /**
     * Записывает в базу знаний статические правила, в частности, описания интересующих сцен девиантного поведения.
     */
    private void initKb(double fps)
    {
        // -Djava.library.path=/usr/lib/swi-prolog/lib/x86_64-linux
        ProcessBuilder builder = new ProcessBuilder("java", "-Xmx" + kbWorkerMemory, "-Xms" + kbWorkerMemory,
                "-Djava.library.path=" + libjplJavaLibraryPath, "-jar", kbWorkerJarPath, generalKbRulesFilePath,
                cameraKbRulesFilePath, Integer.toString(kbHistoryFrames), Integer.toString(devBehIgnorePeriodFrames),
                Double.toString(fps));
        try
        {
            kbWorkerProcess = builder.start();
            kbWorkerIn = new BufferedInputStream(kbWorkerProcess.getInputStream());
            kbWorkerOut = new BufferedOutputStream(kbWorkerProcess.getOutputStream());
            kbWorkerErr = new InputStreamReader(new BufferedInputStream(kbWorkerProcess.getErrorStream()));
        }
        catch (Exception e)
        {
            handleCriticalError("Exception occurred during kb-worker process creation. Manager: " +
                    managerName, e);
            throw new RuntimeException();
        }
        // Читаем ответ, отправляемый kb процессом после успешной инициализации.
        readResponse();
    }


    /**
     * Читает ответ от kb процесса. В случае, если при чтении произойдет ошибка (например, из-за смерти процесса),
     * будет произведена остановка всего pipeline обработки видео, и будет выброшено исключение.
     * Кроме того, остановка pipeline произойдет, если возникло прерывание чтения ответа и выключение менеджера не было
     * инициировано.
     *
     * @return результат обработки видео
     * @throws RuntimeException в случае возникновении ошибки или прерывания
     */
    private KBServiceResponse readResponse() throws RuntimeException
    {
        try
        {
            KBServiceResponse response = KBServiceResponse.parseDelimitedFrom(kbWorkerIn);
            if (response == null)
            {
                // Метод гарантированно выбрасывает исключение
                handleKbWorkerError("The stream was finished. Manager: " + managerName, null);
            }
            return response;
        }
        catch (IOException e)
        {
            if (!getDoStop())
            {
                // Метод гарантированно выбрасывает исключение
                handleKbWorkerError("Exception occurred during reading response. Manager: " +
                        managerName, e);
            }
            throw new RuntimeException();
        }
    }


    /**
     * Отправляет команду kb процессу. В случае, если при чтении произойдет ошибка (например, из-за смерти процесса),
     * будет произведена остановка всего pipeline обработки видео, и будет выброшено исключение.
     *
     * @param command отправляемая команда
     * @throws RuntimeException в случае возникновения ошибки
     */
    private void writeCommand(KBServiceCommand command) throws RuntimeException
    {
        try
        {
            command.writeDelimitedTo(kbWorkerOut);
            kbWorkerOut.flush();
        }
        catch (IOException e)
        {
            // Гарантированно выбрасывает исключение
            handleKbWorkerError("Exception occurred during writing command. Manager: " + managerName, e);
        }
    }


    /**
     * Обрабатывает ошибку при записи команды/чтении ответа kb процесса.
     *
     * @param errorMsg описание ошибки
     * @param e возникшее исключение или null
     * @throws RuntimeException гарантированно будет выброшено данным методом
     */
    private void handleKbWorkerError(String errorMsg, Exception e) throws RuntimeException
    {
        String workerError = null;
        if (!kbWorkerProcess.isAlive())
        {
            // Вероятно, ошибка чтения ответа возникла из-за смерти kb потока. Читаем возможную ошибку.
            try
            {
                workerError = IOUtils.toString(kbWorkerErr);
            }
            catch (IOException ex)
            {
                LOGGER.error("Exception occurred during reading kb worker error stream [{}]", managerName, ex);
            }
        }
        errorMsg = Utils.isEmpty(workerError) ? errorMsg : errorMsg + " -> " + workerError;
        handleCriticalError(errorMsg, e);
        throw new RuntimeException();
    }


    /**
     * Записывает знания, относящиеся к указанному кадру, в базу знаний. Удаляет устаревшие знания, вышедшие за предел
     * рассматриваемого окна времени.
     * Данный метод не вызывается для завершающего кадра с идентификатором -1.
     *
     * @param knowledge знания, извлеченные из кадра
     * @throws RuntimeException в случае ошибки при отправке команды/чтения ответа kb процесса. В случае ошибки также
     * произойдет остановка всего pipeline
     */
    private void addFrameKnowledge(KBEntry knowledge) throws RuntimeException
    {
        writeCommand(KBServiceCommand.newBuilder().setAddFrameKnowledge(knowledge).build());
        // Ожидаем пустой ответ, говорящий об успешной обработке команды
        readResponse();
    }


    /**
     * Осуществляет логический вывод по базе знаний для детектирования девиантного поведения людей.
     */
    private void runInference() throws InterruptedException
    {
        writeCommand(KBServiceCommand.newBuilder().setRunInference(Empty.newBuilder().setEmpty(true)).build());
        KBServiceResponse response = readResponse();
        if (response.getDetections().getEntriesCount() > 0)
        {
            sendDevBehDetectionsToKafka(response.getDetections(), response.getDetections().getFrameId());
        }
    }


    /**
     * Отправляем переданный объект в топик детектированных экземпляров девиантного поведения.
     *
     * @param devBehDetections отправляемый объект
     * @param frameId идентификатор кадра, устанавливаемый в качестве таймштампа сообщения
     * @throws InterruptedException в случае прерывания ожидания подтверждения записи сообщения в топик
     */
    private void sendDevBehDetectionsToKafka(DevBehDetections devBehDetections, long frameId)
            throws InterruptedException
    {
        while (!getDoStop())
        {
            try
            {
                devBehDetectionProducer.send(new ProducerRecord<>(devBehDetectionsTopicPartition.topic(),
                        devBehDetectionsTopicPartition.partition(), frameId, null, devBehDetections)).get();
                break;
            }
            catch (ExecutionException e)
            {
                LOGGER.warn("Exception occurred during sending dev beh detections to Kafka [{}]", managerName);
                sleep(1000);
            }
        }
    }


    @Override
    protected void doLast()
    {
        if (kbWorkerProcess != null)
        {
            kbWorkerProcess.destroyForcibly();
        }

        if (kbStateProducer != null)
        {
            kbStateProducer.close();
        }

        if (devBehDetectionProducer != null)
        {
            devBehDetectionProducer.close();
        }
    }


    @Override
    public String getName()
    {
        return managerName;
    }


    @Override
    protected boolean isDaemon()
    {
        return false;
    }
}
