package com.devbeh.videoproviders;


import java.util.Optional;

import com.devbeh.FrameIdAndBytes;


/**
 * Интерфейс класса, предоставляющего кадры видеозаписи.
 */
public interface VideoProvider
{
    /**
     * Выполняет первый этап инициализации, в результате которого определяется fps.
     * Далее должен быть выполнен второй этап инициализации посредством вызова {@link #initFrameParameters(long, int)}.
     *
     * @throws Exception В случае критичной ошибки, которая приводит к невозможности запуска обработки видео. Ошибки,
     * которые могут иметь временный характер, будут обработаны в реализации провайдера и не будут выброшены наружу.
     * Такие ошибки могут привести только к временной недоступности кадров.
     *
     * @return количество кадров в секунду
     */
    double initFps() throws Exception;

    /**
     * Выполняет второй этап инициализации. Данный метод необходимо вызывать после {@link #initFps()}.
     *
     * @param initialFrameId идентификатор кадра, с которого ожидается видеозапись (включая),
     *                       или {@link Long#MAX_VALUE}, если от провайдера требуется лишь определить fps (кадры при
     *                       этом запрашиваться не будут)
     * @param maxFrameSizePixels максимальный размер кадра видеозаписи, полученный как height * width, в случае
     *                           превышения будет осуществляться ресайз
     */
    void initFrameParameters(long initialFrameId, int maxFrameSizePixels);

    /**
     * Возвращает высоту кадка.
     *
     * @return высота кадра в пикселях
     */
    int getWidth();

    /**
     * Возвращает ширину кадра.
     *
     * @return ширина кадра в пикселях
     */
    int getHeight();

    /**
     * Считывает кадр видеозаписи.
     * frameId ожидается последовательным, но по каким-либо причинам допускается скачек. Кроме того, начальный frameId
     * не обязан быть равен 0. Виртуальное время, к которому будет относиться кадр, будет вычислено
     * как frameId * 1000 / fps (в миллисекундах). В случае, если видеопоток завершен, все последующие вызовы данного
     * метода вернут -1 в качестве идентификатора кадра.
     *
     * @param timeoutMs максимальное время ожидания следующего кадра
     * @return следующий кадр или пустой объект, если вышло время ожидания
     * @throws Exception в случае критичной ошибки, при которой невозможна дальнейшая обработка видео
     */
    Optional<FrameIdAndBytes> read(long timeoutMs) throws Exception;

    /**
     * Закрывает видео поток, освобождает ресурсы.
     *
     * @throws Exception в случае критичной ошибки
     */
    void close() throws Exception;
}
